/*
 Copyright 2015 Nicolas Melot

 This file is part of Pelib-SCC.

 Pelib-SCC is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Pelib-SCC is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Pelib-SCC. If not, see <http://www.gnu.org/licenses/>.

*/

#include <pelib/pelib_scc.h>
#include <RCCE.h>
#include <RCCE_lib.h>

#if 10
#define debug(var) printf("[%s:%s:%d] %s = \"%s\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#define debug_addr(var) printf("[%s:%s:%d] %s = \"%p\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#define debug_int(var) printf("[%s:%s:%d] %s = \"%d\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#define debug_size_t(var) printf("[%s:%s:%d] %s = \"%zu\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#else
#define debug(var)
#define debug_addr(var)
#define debug_int(var)
#define debug_size_t(var)
#endif

const size_t mpb_size = MPB_SIZE;
const char* shared_dir = SHARED_DIR;

void* local_mpb;

struct mpb_stack
{
	void* base_ptr;
	size_t size;
	size_t* stack_ptr;
};
typedef struct mpb_stack mpb_stack_t;

int
pelib_scc_core_id()
{
	return RCCE_ue();
}

int
pelib_scc_core_count()
{
	return RCCE_num_ues();
}

static void* pelib_local_fool_wcb_location = NULL;

int
pelib_scc_init(int *argc, char*** argv)
{
	int retval = RCCE_init(argc, argv);
	if(retval != RCCE_SUCCESS)
	{
		return retval;
	}
	local_mpb = (void*)RCCE_comm_buffer[RCCE_IAM];
	pelib_local_fool_wcb_location = (void*) (RCCE_comm_buffer[RCCE_IAM] - RCCE_LINE_SIZE); // comm buffer is offset by one line, see rcce source for details.
	return 1;
}

int
pelib_scc_finalize()
{
	RCCE_finalize();

	return 1;
}

void*
pelib_scc_malloc_mpb(size_t size)
{
	void* ptr;
	ptr = (void*)RCCE_malloc(size);
	return ptr;
}

int
pelib_scc_free(void* ptr)
{
	RCCE_free((t_vcharp)ptr);
	return 1;
}

#if 0
#define stack_malloc_printf_addr(addr) fprintf(stderr, "[%s:%s:%d] @%s: %08X\n", __FILE__, __FUNCTION__, __LINE__, #addr, addr);
#define stack_malloc_printf_int(addr) fprintf(stderr, "[%s:%s:%d] %s: %d (signed) = %u (unsigned)\n", __FILE__, __FUNCTION__, __LINE__, #addr, addr, addr);
#else
#define stack_malloc_printf_addr(addr)
#define stack_malloc_printf_int(addr)
#endif
mpb_stack_t*
pelib_scc_stack_malloc(size_t size)
{
	mpb_stack_t *stack = (mpb_stack_t*)malloc(sizeof(mpb_stack_t));
	stack_malloc_printf_int(size);
	stack->base_ptr = (void*)RCCE_malloc(size);
	stack_malloc_printf_addr(stack->base_ptr);
	stack->size = size;
	stack->stack_ptr = (size_t*)malloc(sizeof(size_t) * pelib_scc_core_count());
	memset(stack->stack_ptr, 0, sizeof(size_t) * pelib_scc_core_count());
	return stack;
}

#if 0
#define stack_grow_printf_addr(addr) fprintf(stderr, "[%s:%s:%d] @%s: %08X\n", __FILE__, __FUNCTION__, __LINE__, #addr, addr);
#define stack_grow_printf_int(addr) fprintf(stderr, "[%s:%s:%d] %s: %d (signed) = %u (unsigned)\n", __FILE__, __FUNCTION__, __LINE__, #addr, addr, addr);
#else
#define stack_grow_printf_addr(addr)
#define stack_grow_printf_int(addr)
#endif
void*
pelib_scc_stack_grow(mpb_stack_t *stack, size_t size, int id)
{
	stack_grow_printf_int(size);
	stack_grow_printf_int(id);
	stack_grow_printf_addr(stack->base_ptr);
	stack_grow_printf_int(stack->stack_ptr[id]);
	if((size_t)(stack->stack_ptr[id]) + size <= stack->size)
	{
		size_t ptr = stack->stack_ptr[id];
		stack->stack_ptr[id] += size;
		
		return (void*)((size_t)stack->base_ptr + ptr);
	}
	abort();
	return NULL;
}

int
pelib_scc_stack_free(mpb_stack_t* stack)
{
	free(stack->stack_ptr);
	RCCE_free((volatile unsigned char *)stack->base_ptr);
	free(stack);

	return 1;
}

#if 0
#define global_ptr_printf_addr(addr) fprintf(stderr, "[%s:%s:%d] @%s: %08X\n", __FILE__, __FUNCTION__, __LINE__, #addr, addr);
#define global_ptr_printf_int(addr) fprintf(stderr, "[%s:%s:%d] %s: %d (signed) = %u (unsigned)\n", __FILE__, __FUNCTION__, __LINE__, #addr, addr, addr);
#else
#define global_ptr_printf_addr(addr)
#define global_ptr_printf_int(value)
#endif
void*
pelib_scc_global_ptr(void* ptr, int id)
{
	if(id < 0)
	{
		return ptr;
	}
	else
	{
		global_ptr_printf_int(id);
		global_ptr_printf_addr(RCCE_comm_buffer[id]);
		global_ptr_printf_addr(ptr);
		void* global_ptr = (void*)((size_t)RCCE_comm_buffer[id] + (size_t)ptr - (size_t)local_mpb);
		global_ptr_printf_addr(global_ptr);

		return global_ptr;
	}
}

/* Takes a global ptr that belongs to core id and return equivalent offset from local mpb */
/* To be used as buffer allocation comparison purpose, when debugging buffer allocation */
size_t
pelib_scc_local_offset(void* ptr, int id)
{
	if(id < 0) // Treated as local address, no transformation to be performed
	{
		return (size_t)ptr;
	}
	else
	{
		return (size_t)((size_t)ptr - (size_t)RCCE_comm_buffer[id]);
	}
}

size_t
pelib_scc_mpb_size()
{
	return MPB_SIZE;
}

void
pelib_scc_force_wcb()
{
	// *(int*)local_mpb = 1;  //NO! local_mbp is already offsetted by one line. We risk messing up data /johja
	*(int*)pelib_local_fool_wcb_location = 1;
}
void
pelib_scc_global_barrier()
{
	RCCE_barrier(&RCCE_COMM_WORLD);
}

void
pelib_scc_cache_invalidate()
{
	RC_cache_invalidate();
}

