/*
 Copyright 2015 Nicolas Melot

 This file is part of Pelib-SCC.

 Pelib-SCC is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Pelib-SCC is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Pelib-SCC. If not, see <http://www.gnu.org/licenses/>.

*/


/*
 * scc_printf.c
 *
 *  Created on: 17 Apr 2012
 *  Copyright 2012 Nicolas Melot
 *
 * This file is part of pelib-scc.
 *
 *     pelib-scc is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     pelib-scc is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with pelib-scc. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <pelib/sync.h>
#include <pelib/pelib_scc.h>
#include <pelib/scc_printf.h>

#if 10
#define debug(var) printf("[%s:%s:%d:CORE %d] %s = \"%s\"\n", __FILE__, __FUNCTION__, __LINE__, RCCE_ue(), #var, var); fflush(NULL)
#define debug_addr(var) printf("[%s:%s:%d:CORE %d] %s = \"%X\"\n", __FILE__, __FUNCTION__, __LINE__, RCCE_ue(), #var, var); fflush(NULL)
#define debug_int(var) printf("[%s:%s:%d: CORE %d] %s = \"%d\"\n", __FILE__, __FUNCTION__, __LINE__, RCCE_ue(), #var, var); fflush(NULL)
#define debug_size_t(var) printf("[%s:%s:%d: CORE %d] %s = \"%zu\"\n", __FILE__, __FUNCTION__, __LINE__, RCCE_ue(), #var, var); fflush(NULL)
#else
#define debug(var)
#define debug_addr(var)
#define debug_int(var)
#define debug_size_t(var)
#endif

pelib_scc_flag critical_flag;
FILE *stdout_backup = NULL, *stderr_backup = NULL;

static void
start_critical_section(pelib_scc_flag *critical_flag)
{
	pelib_scc_flag_alloc(critical_flag);
	pelib_scc_flag_write(critical_flag, PELIB_SCC_FLAG_UNSET, RCCE_ue());

	RCCE_barrier(&RCCE_COMM_WORLD);

	if (RCCE_ue() > 0)
	{
		pelib_scc_wait_until(*critical_flag, PELIB_SCC_FLAG_SET);
	}
}

static void
stop_critical_section(pelib_scc_flag *critical_flag)
{
	if (RCCE_ue() < RCCE_num_ues() - 1)
	{
		pelib_scc_flag_write(critical_flag, PELIB_SCC_FLAG_SET, RCCE_ue() + 1);
	}

	RCCE_barrier(&RCCE_COMM_WORLD);
	pelib_scc_flag_free(critical_flag);
}

//static FILE *out, *err, *redir_out, *redir_err;
static int bak_out, bak_err, redir_out, redir_err;

void
pelib_scc_init_redirect()
{
        char file_out[sizeof(SHARED_DIR) + 1 + sizeof(PELIB_SCC_PRINTF_OUT) + 2];
        char file_err[sizeof(SHARED_DIR) + 1 + sizeof(PELIB_SCC_PRINTF_ERR) + 2];

        umask(0111);
        sprintf(file_out, "%s/%s%02i", shared_dir, PELIB_SCC_PRINTF_OUT, RCCE_ue());
        sprintf(file_err, "%s/%s%02i", shared_dir, PELIB_SCC_PRINTF_ERR, RCCE_ue());

        redir_out = open(file_out, O_CREAT | O_TRUNC | O_RDWR, 0666);
	if(redir_out == -1)
	{
                fprintf(stderr, "[CORE %d] Could not open standard output file: %s\n", RCCE_ue(), file_out);
                abort();
	}
        redir_err = open(file_err, O_CREAT | O_TRUNC | O_RDWR, 0666);
	if(redir_err == -1)
	{
                fprintf(stderr, "[CORE %d] Could not open error output file: %s\n", RCCE_ue(), file_err);
                abort();
	}
}

void
pelib_scc_finalize_redirect()
{
	close(bak_out);
	close(bak_err);
}

void
pelib_scc_set_redirect()
{
	fflush(stdout);
	fflush(stderr);

	bak_out = dup(STDOUT_FILENO);
	bak_err = dup(STDERR_FILENO);

	dup2(redir_out, STDOUT_FILENO);
	dup2(redir_err, STDERR_FILENO);

	close(redir_out);
	close(redir_err);
}

void
pelib_scc_stop_redirect()
{
	dup2(bak_out, STDOUT_FILENO);
	dup2(bak_err, STDERR_FILENO);
}

